/**************************************************************************************

        MUSTUXLIB - THE COMMON LIBRARY FOR ALL MUSTUX APPLICATIONS
        AUTHOR : See AUTHORS file for details

        This software is distributed under the terms of the GNU General Public License
        as specified in the COPYING file.

***************************************************************************************/

#include <stdio.h>
#include <qtimer.h>
#include <stdlib.h>

#include "MustuxJogMouseBoard.hh"
#include "MustuxJogMouseBoardAction.hh"
#include "MustuxJogMouseBoardMessage.hh"
#include "MustuxDebugger.hh"
#include <qtextstream.h>
#include <qfile.h>


#define SKIP do fread(&c, 1, 1, mapFile); while (((c==0x20) || (c==0x09)) && (!feof(mapFile)));

MustuxJogMouseBoard::MustuxJogMouseBoard(MustuxInterface* pInterface) :  MustuxEngine( pInterface)
        {
        PENTERCONS;
        assocInterface = pInterface;
        clearTime = 2000;
        holdSensitivenessFactor=15; // maybe this and next can be integrated into a single factor
        holdcheckInterval=10;
        doublePressQuitTime = 180;
        collectedDigit=0;
        collectedNumber = -1;
        sCollectedNumber = "-1";
        locked=false; // THIS SHOULD CALL everythingLocked        
        PEXITCONS;
        }

MustuxJogMouseBoard::~MustuxJogMouseBoard()
        {
        PENTERDES;
        PEXITDES;
        }


int MustuxJogMouseBoard::init(QString mapFilename)
        {
        PENTER;
        activate();
        clearOutputTimer=0;
        reset();
        init_map(mapFilename);
        holdTimer =  new QTimer(this);
        connect( holdTimer, SIGNAL(timeout()), this, SLOT(holding()) );
        secondChanceTimer = new QTimer(this);
        connect( secondChanceTimer, SIGNAL(timeout()), this, SLOT(quit_second_chance()) );
        clearOutputTimer = new QTimer();
        connect( clearOutputTimer, SIGNAL(timeout()), this, SLOT(clear_output()) );
        PEXIT;
        return 1;
        }


void MustuxJogMouseBoard::activate()
        {
        assocInterface->defaultLcd->clear_line(1);
        assocInterface->defaultLcd->clear_line(2);
        assocInterface->defaultLcd->clear_line(3);
        assocInterface->defaultLcd->clear_line(4);
        assocInterface->defaultLcd->print("JMB ACTIVE",4);
        isFirstPress=true;
        active=true;
        }


void MustuxJogMouseBoard::suspend()
        {
        assocInterface->defaultLcd->print("JMB SUSPENDED",4);
        active=false;
        }


int MustuxJogMouseBoard::finalize()
        {
        delete holdTimer;
        delete secondChanceTimer;
        delete clearOutputTimer;
        return 1;
        }


// ----------------------- JMB ENGINE : EVENT LEVEL HANDLING ---------------------

// reset the event stack and the press 'stack' (not exactly a stack)
void MustuxJogMouseBoard::reset()
        {
        isFirstPress = true;
        isHolding = false;
        isTimerRunning = false;
        isKeyPressLocked = false;
        stackIndex = 0;
        keyPressCounter = 0;
        holdTime=0;
        press1_k1 = 0;
        press1_k2 = 0;
        press2_k1 = 0;
        press2_k2 = 0;
        wholeAction = -1;

        for (int i=0; i < STACK_SIZE; i++)
                {
                eventType[i] = 0;
                eventStack[i] = 0;
                time[i] = 0;
                }
        }


void MustuxJogMouseBoard::keyPressEvent ( QKeyEvent* e )
        {
        if (active)
                {
                if (!e->isAutoRepeat())
                        catch_press(e);
                }
        e->ignore();
        }

void MustuxJogMouseBoard::keyReleaseEvent ( QKeyEvent* e )
        {
        if (active)
                {
                if (!e->isAutoRepeat())
                        catch_release(e);
                }
        e->ignore();
        }


// Everthing starts here. Catch event takes anything happen in the keyborad and pushes it
// into a stack.
int MustuxJogMouseBoard::catch_press( QKeyEvent* e)
        {
        int k = uppercase(e->key());
        if (!isKeyPressLocked)
                {
                if (keyPressCounter < 2)
                        {
                        keyPressCounter++;
                        push_event(JMB_KEY_PRESS, k);
                        if (!isTimerRunning)
                                {
                                holdTime = 0;
                                isTimerRunning = true;
                                holdTimer->start( holdcheckInterval );
                                }
                        }
                else
                        {
                        isKeyPressLocked = true;
                        }
                }
        return 1;
        }

int MustuxJogMouseBoard::catch_release( QKeyEvent* e)
        {

        int k = uppercase(e->key());
        if (!is_fake(k))
                {
                push_event(JMB_KEY_RELEASE, k);
                release_checker();
                }
        
        return 1;
        }


int MustuxJogMouseBoard::uppercase(int k)
        {
        /*cout << "k=" << k << " , char(k)=" << char(k) << endl;
        if ((k>90) && (k<106))
                k-=32;
        */
        return k;
        }

// this is a intermediate step to push. Only non-fake events are allowed to be pushed
bool MustuxJogMouseBoard::is_fake(int keyval)
        {
        // A fake event is an event which is generated by a ignored JMB_KEY_PRESS
        bool b =     ( keyval!=eventStack[0] )
                && ( keyval!=eventStack[1] )
                && ( keyval!=eventStack[2] )
                && ( keyval!=eventStack[3] );
        return b;
        }


// This pushed an event to the stack
void MustuxJogMouseBoard::push_event(  int pType,  int pKey )
        {
        //PMESG("Pushing EVENT " << ( pType==JMB_KEY_PRESS ? "PRESS" : "RELEASE" ) << " Key="<< pKey)
        if (stackIndex < STACK_SIZE)
                {
                eventType[stackIndex] = pType;
                eventStack[stackIndex] = pKey;
                stackIndex++;
                }
        }


// This is the one who consolidate a press. It can detect
// if the press is K or KK, and also calls the push_press handler
void MustuxJogMouseBoard::release_checker()
        {

        if (isHolding)
                {
                finish_hold();
                return;
                }
        // If it is not fake, then lets take a look in the pairs
        // Remember that eventType[0]==JMB_KEY_PRESS always
        if (eventType[1]==JMB_KEY_RELEASE)
                {
                // event 0 and event 1 are a pair because event 1 must be the release of 0
                // since the event 1 was a release, then the action is finished
                // Now we must push_press, but only if it is a release very fast (that
                // didnt cause a HOLD.
                if (!isHolding)
                        push_press(eventStack[0],0);

                return;
                }
        // action 1 is ALSO a JMB_KEY_PRESS. So we have a Double_Key
        // Now we dont know if we are in action 2 (stackIndex==2)
        // or in the action 3 (stackIndex==3)
        if ((stackIndex-1)==2)
                {
                // So we are processing the action 2, which is a release
                // So action 2 can be release of action 0 or 1 , so..
                if (eventStack[2]==eventStack[0])
                pairOf2 = 0;
                else
                pairOf2 = 1;
                // the event is not finished yet but it must be avoided
                // any reentrance (any new JMB_KEY_PRESS)
                isKeyPressLocked = true;
                
                return;
                }
        // we are in the action 3 . the last!
        // the event is finished , surely. But we must check the pairs
        // we already know who action 2 is pair of. So ...
        if (pairOf2 == 0)
                pairOf3 = 1;
        else
                pairOf3 = 0;
        // The event is finished, and we know the pairs. So..
        // I must push_press, but only if it is a release very fast (that
        // didnt cause a HOLD.
        if (!isHolding)
                push_press(eventStack[0],eventStack[1]);
        
        }


void MustuxJogMouseBoard::holding()
        {
        holdTime++;
        if (holdTime > holdSensitivenessFactor)
                {
                //PMESG("holdTimer timeout. Dispatching hold");
                isHolding = true;
                holdTimer->stop();
                dispatch_hold();
                }
        else
                {
                //PMESG("Time left to consider a hold " << holdSensitivenessFactor-holdTime << "ms");
                }
        }

// ----------------------------- JMB ENGINE : PRESS LEVEL HANDLING -------------------------


// This guy pushes a press (the first and the second)
void MustuxJogMouseBoard::push_press(int k1,int k2)
        {
        //PMESG("Pushing ACTION PRESS : k1=" << k1 << " k2=" << k2);
        assocInterface->defaultLcd->clear_line(1);
        holdTimer->stop();
        if (isFirstPress)
                {
                // this is the first press
                //PMESG("First press detected !");
                // first try to find some action like k1k2
                press1_k1 = k1;
                press1_k2 = k2;

                // first check if it is a number collection
                check_number_collection();
                if (isCollecting)
                        {
                        // another digit was collected. Fine , go on jmb job..
                        reset();
                        
                        return;
                        }


                int action = identify_first_press();
                //PMESG("Got " << action << " from identify_first_press()");
                if (action < 0)
                        {
                        // Action not identified. Maybe is part of a Double press action. so...
                        give_a_chance_for_second_press();
                        
                        return;
                        }
                // action exists !! now must check if is not an immediate action
                if (!map[action].isInstantaneous)
                        {
                        //PMESG("Although this could be an SINGLE PRESS Action, it is not protected, so...");
                        // action is not an immediate action, so...
                        give_a_chance_for_second_press();

                        return;
                        }

                // Action exists AND it is a immediate action. So
                // forces it to be a SinglePressAction
                //PMESG("This is immediate action. Giving immediate response");
                //isDoublePress = false;
                dispatch_action(action);
                conclusion();
                }
        else
                {
                secondChanceTimer->stop();
                press2_k1 = k1;
                press2_k2 = k2;
                if (press2_k1!=0)
                        {
                        // this is the second press
                        //PMESG("Second press detected !");
                        }

                // try to complement the first press.
                wholeAction = complement_first_press();
                //PMESG("Got " << wholeAction << " from complement_first_press()");
                if (wholeAction >= 0)
                        {
                        //PMESG("Action " << wholeAction << " recognized. Dispatching...");
                        if (locked)
                                {
                                PMESG("OLA!");
                                assocInterface->defaultLcd->print("Jog Mouse Board is LOCKED !",0);
                                return;
                                }

                        dispatch_action(wholeAction);
                        }
                else
                        {
                        assocInterface->defaultLcd->print("Unknown <K> Action",0);
                        }

                conclusion();
                }

        }





int MustuxJogMouseBoard::identify_first_press()
        {
        press1Type = 0;
        // First we need to know the first press action type.
        if (press1_k2==0) // {K}
                {
                if (!isHolding)
                        {
                        //PMESG("Detected <K>");
                        press1Type = FKEY;
                        }
                else
                        {
                        //PMESG("Detected [K]");
                        press1Type = HKEY;
                        }
                }
        else // {KK}
                {
                if (!isHolding)
                        {
                        //PMESG("Detected <KK>");
                        press1Type = FKEY2;
                        }
                else
                        {
                        //PMESG("Detected [KK]");
                        press1Type = HKEY2;
                        }
                }

        // Press 1 Type identified .
        // Now we ask : is there any action like {k1k2}/{0 0} ?
        //PMESG("Now searching for a match SINGLE PRESS action in map. ( mapSize=" << mapSize << " )");
        for (int i=0; i<mapSize; i++)
                        {
                        int ap1k1 = map[i].press1_key1;
                        int ap1k2 = map[i].press1_key2;
                        int ap2k1 = map[i].press2_key1;
                        int ap2k2 = map[i].press2_key2;

                        //PMESG("COMPARING " <<  ap1k1     << "," << ap1k2     << "," << ap2k1 << "," << ap2k2 <<   " WITH "    <<  press1_k1 << "," << press1_k2 << ",0,0")

                        if (map[i].type != press1Type ) continue;


                        if (
                                (
                                ((ap1k1==press1_k1) && (ap1k2==press1_k2))
                                ||
                                ((ap1k1==press1_k2) && (ap1k2==press1_k1))
                                )
                                &&
                                ( ap2k1 == 0 )
                                &&
                                ( ap2k2 == 0 )
                        )
                                {
                                //PMESG("Found a match in i=" << i);
                                // 'i' is a candidate for first press
                                return i;
                                }
                        }
        // No candidate found.
        return -1;
        }





int MustuxJogMouseBoard::complement_first_press()
        {
        //PMESG("Adding the pair press2_k1=" << press2_k1 << " press2_k2=" << press2_k2 << " to press1_k1=" << press1_k1 << ",press1_k2=" << press1_k2);
        int wholeActionType;
        if (press2_k1!=0)
                {
                if (!isHolding)
                        {
                        if (press1_k2==0) // first press is <K> (I know that its not [K] because if it was I'd never reach complement_first_press())
                                {
                                if (press2_k2==0)
                                        if (press1_k1 == press2_k1)
                                                {
                                                //PMESG("Whole action is a <<K>>");
                                                wholeActionType = D_FKEY;  // <<K>>
                                                }
                                        else
                                                {
                                                //PMESG("Whole action is a >K>K");
                                                wholeActionType = S_FKEY_FKEY; // >K>K
                                                }
                                else
                                        {
                                        //PMESG("Whole action is a >K>KK");
                                        wholeActionType = S_FKEY_FKEY2;  // >K>KK
                                        }
                                }

                        else
                                {
                                if (press2_k2==0)
                                        {
                                        //PMESG("Whole action is a >KK>K");
                                        wholeActionType = S_FKEY2_FKEY;  // >KK>K
                                        }
                                else
                                        {
                                        if (
                                        ((press1_k1==press2_k1) && (press1_k2==press2_k2)) ||
                                        ((press1_k1==press2_k2) && (press1_k2==press2_k1))
                                        )
                                                {
                                                //PMESG("Whole action is a <<KK>>");
                                                wholeActionType = D_FKEY2;
                                                }
                                        else
                                                {
                                                //PMESG("Whole action is a >KK>KK");
                                                wholeActionType = S_FKEY2_FKEY2;
                                                }
                                        }
                                }
                        }
                else
                        {
                        if (press1_k2==0) // first press is <K> (I know that its not [K] because if it was I'd never reach complement_first_press())
                                {
                                if (press2_k2==0)
                                        if (press1_k1 == press2_k1)
                                                {
                                                //PMESG("Whole action is a <[K]>");
                                                wholeActionType = FHKEY;
                                                }
                                        else
                                                {
                                                //PMESG("Whole action is a >K[K]");
                                                wholeActionType = S_FKEY_HKEY;
                                                }
                                else
                                        {
                                        //PMESG("Whole action is a >K[KK]");
                                        wholeActionType = S_FKEY_HKEY2;
                                        }
                                }
                        else
                                {
                                if (press2_k2==0)
                                        {
                                        //PMESG("Whole action is a KK[K]");
                                        wholeActionType = S_FKEY2_HKEY;
                                        }
                                else
                                        {
                                        //PMESG("Whole action is a >KK[KK]");
                                        wholeActionType = S_FKEY2_HKEY2;
                                        }
                                }
                        }
                }
        else
                {
                //PMESG("Second Press is totally null. Assuming wholeActionType = press1Type (" << press1Type << ")");
                wholeActionType = press1Type;
                }
        // whole action type identified .

        //PMESG("Now searching for a match action...");
        for (int i=0; i<mapSize; i++)
                        {
                        if ( map[i].type != wholeActionType ) continue;
                        int ap1k1 = map[i].press1_key1;
                        int ap1k2 = map[i].press1_key2;
                        int ap2k1 = map[i].press2_key1;
                        int ap2k2 = map[i].press2_key2;
                        //PMESG("COMPARING " <<  ap1k1     << "," << ap1k2     << "," << ap2k1 << "," << ap2k2 <<  " WITH "    <<  press1_k1 << "," << press1_k2 << ",0,0")
                        if (
                                (
                                ((ap1k1==press1_k1) && (ap1k2==press1_k2))
                                ||
                                ((ap1k1==press1_k2) && (ap1k2==press1_k1))
                                )
                                &&
                                (
                                ((ap2k1==press2_k1) && (ap2k2==press2_k2))
                                ||
                                ((ap2k1==press2_k2) && (ap2k2==press2_k1))
                                )
                        )
                                {
                                //PMESG("Found a match in i=" << i);
                                // 'i' is a candidate the whole action
                                return i;
                                }
                        }
        // No candidate found.
        return -1;
        }


void MustuxJogMouseBoard::give_a_chance_for_second_press()
        {

        //PMESG("Waiting " <<  doublePressQuitTime << "ms for second press...");
        secondChanceTimer->start( doublePressQuitTime );
        isFirstPress=false;

        }

void MustuxJogMouseBoard::quit_second_chance()
        {
        
        secondChanceTimer->stop();
        //PMESG("Im sorry, you lost your chance for second press...");
        push_press(0 , 0); // no second press performed, so I am adding a pair of Zeros as second press and keep going...
        
        }


// ----------------------------- JMB ENGINE : ACTION LEVEL HANDLING -------------------------
void MustuxJogMouseBoard::dispatch_action(int action)
        {
        assocInterface->defaultLcd->print(map[action].name,0);
        PMESG("Processing Action %d ( %s ) ",action, (const char*) map[action].name.latin1() );

        MustuxJogMouseBoardMessage* m =
                        new MustuxJogMouseBoardMessage(
                                                        map[action].name,
                                                        false,
                                                        action,
                                                        (QWidget*) 0,
                                                        0,
                                                        0,
                                                        get_action_parameter(action),
                                                        get_collected_number()
                                                        );
        assocInterface->process_jmb_action(m);
        delete m;
        }



// This is called by
void MustuxJogMouseBoard::dispatch_hold()
        {
        if (clearOutputTimer!=0)
                {
                clearOutputTimer->stop();
                isHoldingOutput=false;
                }
        assocInterface->defaultLcd->clear_line(1);

        wholeAction = -1;
        if (isFirstPress)
                {
                press1_k1 = eventStack[0];
                press1_k2 = eventStack[1];
                wholeAction = identify_first_press(); // I can consider first press the last because there is nothing after a []
                }
        else
                {
                press2_k1 = eventStack[0];
                press2_k2 = eventStack[1];
                wholeAction = complement_first_press();
                }

        if (wholeAction>=0)
                {
                PMESG("Processing HOLD Action %d ( %s ) ",wholeAction,(const char*) map[wholeAction].name.latin1());
                MustuxJogMouseBoardAction jmbe = map[wholeAction];
                assocInterface->defaultLcd->print(jmbe.name,0);
                if ((jmbe.useX) && (jmbe.useY))
                        {
                        assocInterface->defaultLcd->print("Move Mouse U/D/L/R",1);
                        }
                else if (jmbe.useX)
                        {
                        assocInterface->defaultLcd->print("Move Mouse Left/Right",1);
                        }
                else if(jmbe.useY)
                        {
                        assocInterface->defaultLcd->print("Move Mouse Up/Down",1);
                        }
                //PMESG("Processing hold action " << wholeAction );
                MustuxJogMouseBoardMessage* m = new MustuxJogMouseBoardMessage(
                                                        map[wholeAction].name,
                                                        true,
                                                        wholeAction,
                                                        (QWidget*) 0,
                                                        0,
                                                        0,
                                                        get_action_parameter(wholeAction),
                                                        get_collected_number()
                                                        );

                assocInterface->process_jmb_action(m);
                }
        else
                {
                assocInterface->defaultLcd->print("Unknown [K] Action !",0);
                }
        stop_collecting();
        // note that we dont call conclusion here :)

        }




void MustuxJogMouseBoard::finish_hold()
        {
        assocInterface->defaultLcd->clear_line(1);
        //PMESG("Finishing hold action " << wholeAction);
        if (wholeAction>=0)
                {
                MustuxJogMouseBoardMessage* m =
                                new MustuxJogMouseBoardMessage(
                                                                map[wholeAction].name,
                                                                false,
                                                                wholeAction,
                                                                (QWidget*) 0,
                                                                0,
                                                                0,
                                                                get_action_parameter(wholeAction),
                                                                get_collected_number()
                                                                );

                assocInterface->process_jmb_action(m);
                }
        conclusion();
        }


void MustuxJogMouseBoard::conclusion()
        {
        assocInterface->defaultLcd->print("Action Finished",4);
        reset();
        hold_output();
        }





void MustuxJogMouseBoard::hold_output()
        {
        if (!isHoldingOutput)
                {
                clearOutputTimer->start(clearTime);
                isHoldingOutput=true;
                }
        }



int MustuxJogMouseBoard::init_map(QString mapFilename)
        {
        PENTER;
        mapSize=0;
        QFile mapFile(mapFilename);
        if ( mapFile.open(IO_ReadOnly) )
                {
                PMESG("INITIALIZING JMB MAP ... ");
                int index=0;

                QTextStream t( &mapFile );        // use a text stream
                QString line;
                while ( !t.eof() )
                        {
                        line = t.readLine();
                        int x = line.find("#");
                        if (x>=0) line = line.left(x);
                        line = line.simplifyWhiteSpace();
                        if (line.length()==0)
                                continue;

                        bool useX=false,useY=false;
                        int actionTypeCode=0, nKeys=0;
                        int keyCode[4];
                        int parameter;

                        int x2=line.find("\"",1);
                        QString label = line.mid(1,x2-1);
                        line=line.mid(x2+2);
                        int x3=line.find(" ");
                        QString actionType = line.left(x3);
                        line=line.mid(x3+1);

                        if      ( strcmp(actionType,"<K>")    == 0 )  { actionTypeCode = FKEY;          nKeys = 1; }
                        else if ( strcmp(actionType,"<KK>")   == 0 )  { actionTypeCode = FKEY2;         nKeys = 2; }
                        else if ( strcmp(actionType,"[K]")    == 0 )  { actionTypeCode = HKEY;          nKeys = 1; }
                        else if ( strcmp(actionType,"[KK]")   == 0 )  { actionTypeCode = HKEY2;         nKeys = 2; }
                        else if ( strcmp(actionType,"<<K>>")  == 0 )  { actionTypeCode = D_FKEY;        nKeys = 1; }
                        else if ( strcmp(actionType,"<<KK>>") == 0 )  { actionTypeCode = D_FKEY2;       nKeys = 2; }
                        else if ( strcmp(actionType,"<[K]>")  == 0 )  { actionTypeCode = FHKEY;         nKeys = 1; }
                        else if ( strcmp(actionType,"<[KK]>") == 0 )  { actionTypeCode = FHKEY2;        nKeys = 2; }
                        else if ( strcmp(actionType,">K>K")   == 0 )  { actionTypeCode = S_FKEY_FKEY;   nKeys = 2; }
                        else if ( strcmp(actionType,">K>KK")  == 0 )  { actionTypeCode = S_FKEY_FKEY2;  nKeys = 3; }
                        else if ( strcmp(actionType,">KK>K")  == 0 )  { actionTypeCode = S_FKEY2_FKEY;  nKeys = 3; }
                        else if ( strcmp(actionType,">KK>KK") == 0 )  { actionTypeCode = S_FKEY2_FKEY2; nKeys = 4; }
                        else if ( strcmp(actionType,">K[K]")  == 0 )  { actionTypeCode = S_FKEY_HKEY;   nKeys = 2; }
                        else if ( strcmp(actionType,">K[KK]") == 0 )  { actionTypeCode = S_FKEY_HKEY2;  nKeys = 3; }
                        else if ( strcmp(actionType,">KK[K]") == 0 )  { actionTypeCode = S_FKEY2_HKEY;  nKeys = 3; }
                        else if ( strcmp(actionType,">KK[KK]")== 0 )  { actionTypeCode = S_FKEY2_HKEY2; nKeys = 4; }

                        int x4=line.find(" ");
                        QString sUseX = line.left(x4);
                        if (sUseX=="0") useX=false; else useX=true;
                        line=line.mid(x4+1);

                        int x5=line.find(" ");
                        QString sUseY = line.left(x5);
                        if (sUseX=="0") useY=false; else useY=true;
                        line=line.mid(x5+1);

                        // get nKeys hex codes
                        for (int nk=0; nk<4; nk++) keyCode[nk]=0;
                        for (int nk=0; nk<nKeys; nk++)
                                {
                                int x6=line.find(" ");
                                QString sTag = line.left(x6);
                                line=line.mid(x6+1);
                                int keycode=strtol((char*)sTag.latin1(),0,16);
                                keyCode[nk]=keycode;
                                }

                        // get the parameter
                        QString sParameter = line;
                        parameter = sParameter.toInt();

                        // Fix the keyCode positions
                        if      (( actionTypeCode == D_FKEY  )     ||
                                ( actionTypeCode == FHKEY  ))        { keyCode[2]=keyCode[0]; }
                        else if (( actionTypeCode == S_FKEY_FKEY ) ||
                                ( actionTypeCode == S_FKEY_HKEY ))   { keyCode[2]=keyCode[1]; keyCode[1]=0; }
                        else if (( actionTypeCode == D_FKEY2 )     ||
                                ( actionTypeCode == FHKEY  ))        { keyCode[2]=keyCode[0]; keyCode[3]=keyCode[1]; }

                        PMESG("ADDED action: %s : %s  keys=%d,%d,%d,%d  par=%d", (const char*) label.latin1(),  (const char*)  actionType.latin1(), keyCode[0],keyCode[1],keyCode[2],keyCode[3], parameter);
                        map[index++].set(label, actionTypeCode, keyCode[0], keyCode[1], keyCode[2], keyCode[3], useX, useY, false, parameter);

                        }

                mapSize=index;
                PMESG("Optimizing map for best performance ...");
                int optimizedActions=0;
                for (int i=0; i<mapSize; i++)
                        {
                        int c1A=map[i].press1_key1;
                        int c2A=map[i].press1_key2;

                        if ( (map[i].type==FKEY) )
                                {
                                bool alone = true;
                                for (int j = 0; j < mapSize; j++)
                                        {
                                        if (j==i) continue;
                                        int tt    = map[j].type;
                                        int t1A   = map[j].press1_key1;
                                        if  (
                                                (t1A==c1A)  &&
                                                (
                                                        (tt==D_FKEY)       ||
                                                        (tt==FHKEY)        ||
                                                        (tt==S_FKEY_FKEY)  ||
                                                        (tt==S_FKEY_FKEY2) ||
                                                        (tt==S_FKEY_HKEY)  ||
                                                        (tt==S_FKEY_HKEY2)
                                                )
                                            )
                                        alone=false;
                                        }
                                if (alone)
                                        {
                                        PMESG("Setting <K> action %s as protected (Instantaneous response)",(const char*) map[i].name.latin1());
                                        map[i].set_instantaneous(true);
                                        optimizedActions++;
                                        }
                                else
                                        map[i].set_instantaneous(false);

                                }
                        else if ((map[i].type==FKEY2))
                                {
                                bool alone = true;
                                for (int j = 0; j < mapSize; j++)
                                        {
                                        if (j==i) continue;
                                        int tt    = map[j].type;
                                        int t1A   = map[j].press1_key1;
                                        int t2A   = map[j].press1_key2;
                                        if (
                                        ((t1A==c1A) && (t2A==c2A)) &&
                                        (
                                        (tt==HKEY)         ||
                                        (tt==D_FKEY2)      ||
                                        (tt==S_FKEY2_FKEY) ||
                                        (tt==S_FKEY2_FKEY2)||
                                        (tt==S_FKEY2_HKEY) ||
                                        (tt==S_FKEY2_HKEY2)
                                        )
                                        )
                                        alone=false;
                                        }
                                if (alone)
                                        {
                                        //PMESG("Setting <KK> action " << map[i].name << " for instantaneous response");
                                        map[i].set_instantaneous(true);
                                        optimizedActions++;
                                        }
                                else
                                        map[i].set_instantaneous(false);
                                }
                        }
                PMESG("JMB MAP INITIALIZED! %d actions registered ( %d optimized) .", mapSize, optimizedActions);
                mapFile.close();
                }
        else
                {
                PERROR("Cannot open jmb.map");
                }

        PEXIT;
        return 1;
        }






// callbacks


void MustuxJogMouseBoard::clear_output()
        {

        if ((isHoldingOutput))
                {
                // UNCOMMENT ME ASAP output->clear();
                clearOutputTimer->stop();
                isHoldingOutput=false;
                // UNCOMMENT ME ASAP output->draw_all_grids();
                for (int k=0; k<4; k++)
                        {
                        assocInterface->defaultLcd->clear_line(k);
                        assocInterface->defaultLcd->draw_line_grid(k);
                        }
                assocInterface->defaultLcd->print("Protux JMB Active",4);
                }

        }


// Settings

void MustuxJogMouseBoard::set_clear_time(int time)
        {
        clearTime = time;
        }

void MustuxJogMouseBoard::set_hold_sensitiveness(int factor)
        {
        holdSensitivenessFactor=factor;
        }

void MustuxJogMouseBoard::set_double_press_interval(int time)
        {
        doublePressQuitTime = time;
        }


QString MustuxJogMouseBoard::get_action_name(int a)
        {
        return map[a].name;
        }

int MustuxJogMouseBoard::get_action_parameter(int a)
        {
        return map[a].parameter;
        }

// Number colector

void MustuxJogMouseBoard::check_number_collection()
        {
        if ((press1_k1 >= 0x30) && (press1_k1 <= 0x39))
                {
                if (!isCollecting)
                        {
                        PMESG("Starting number collection...");
                        sCollectedNumber="";
                        }
                isCollecting = true;
                collectedDigit++;
                sCollectedNumber.append( QChar(press1_k1) ); // it had a ",1" complement after press1_k1... why?
                //PMESG("Collected " << sCollectedNumber << " so far...") ;
                QString sn = "NUMBER " + sCollectedNumber;
                collectedNumber = atoi(sCollectedNumber);
                assocInterface->defaultLcd->print(sn,3);
                }
        else
                stop_collecting();

        }

int MustuxJogMouseBoard::get_collected_number()
        {
        int n = collectedNumber ;
        collectedNumber = -1; // collectedNumber has a life of only one get.
        return n;
        }

void MustuxJogMouseBoard::stop_collecting()
        {
        isCollecting=false;
        collectedDigit = 0;
        collectedNumber = atoi(sCollectedNumber);
        sCollectedNumber ="-1";
        assocInterface->defaultLcd->print("Number Collection Finished",3);
        }


void MustuxJogMouseBoard::lock()
        {
        PENTER;
        locked=true;
        PMESG("now locked=%d",locked);
        PEXIT;
        }

void MustuxJogMouseBoard::unlock()
        {
        PENTER;
        locked=false;
        PMESG("now locked=%d",locked);
        PEXIT;
        }

bool MustuxJogMouseBoard::is_locked()
        {
        return locked;
        }
//eof
